import asyncio
from collections import defaultdict
from typing import List, Any
from dataclasses import replace

from deepdiff import DeepHash

from .typedefs import Request, Response, Batch, BatchRequestModel


class AsyncDeduplicator(BatchRequestModel):
    def __init__(self, model: BatchRequestModel, correctness: bool = False):
        # Actual API
        self.model = model
        self.correctness = correctness # Happy to remove this if not needed

    async def batch_request(self, batch: Batch) -> List[Response]:
        # find duplicate prompts
        key2request_and_namespace = {}
        key2requests = defaultdict(list)

        # requests with the same prompt and namespace are identical
        # prompt&namespace are hashed to a key, then all requests under this key are bundled together
        for request in batch.requests:
            request_and_namespace = (
                request.hash(),
                request.namespace,
                "_".join(request.request_id.split("_")[:-1]) if self.correctness else None
                )
            key = DeepHash(request_and_namespace)[request_and_namespace]

            key2request_and_namespace[key] = request_and_namespace
            key2requests[key].append(request)

        # bundle requests and send them as one
        merged_requests = []
        for key in key2request_and_namespace:
            request_and_namespace = key2request_and_namespace[key]
            requests_to_merge = key2requests[key]

            # first entry in metadata is the number of desired answers
            total = sum([m.n for m in requests_to_merge])

            # we now dispatch all identical prompts in one request
            merged_request = replace(requests_to_merge[0], n=total)
            merged_requests.append(merged_request)

        responses = await self.model.batch_request(
            Batch(requests=merged_requests))

        # map the results back to the keys
        key2results = {key: result for key, result in zip(
            key2request_and_namespace.keys(), responses)}

        # Flags whether the result was duplicated or not
        for key, result in key2results.items():
            if result.cached:
                flags = [not flag for flag in result.cached] # reversed cached flags
                # The first duplicated result that is not cached is the original.
                # The remaining uncached results are the duplicates.
                for i, flag in enumerate(flags):
                    if flag:
                        flags[i] = not flags[i]
                        break
            else:
                flags = [False] + [True] * (len(result.data) - 1)
            key2results[key] = key2results[key]._replace(duplicated=flags)

        # split the results back into the individual requests
        results = []
        for request in batch.requests:
            request_and_namespace = (
                request.hash(),
                request.namespace,
                "_".join(request.request_id.split("_")[:-1]) if self.correctness else None
                )
            key = DeepHash(request_and_namespace)[request_and_namespace]

            # our API always works with lists of requests and then for each request a list of responses
            # correspondingly this is a list of lists
            # but this result corresponds to the one deduplicated request (with many responses)
            # so it should be a list of length one
            result = key2results[key]
            metadata = key2requests[key].pop(0)

            assert metadata.n == request.n
            assert len(result.data) >= metadata.n

            results.append(
                Response(
                    data=result.data[:metadata.n],
                    cached=result.cached[:metadata.n] if result.cached else [False]*len(result.data[:metadata.n]),
                    duplicated = result.duplicated[:metadata.n]
                )
            )

            key2results[key]= Response(
                data=result.data[metadata.n:],
                cached=result.cached[metadata.n:] if result.cached else [False]*len(result.data[metadata.n:]),
                duplicated=result.duplicated[metadata.n:]
            )
        return results



from typing import NamedTuple, List, Protocol, Any, runtime_checkable, Optional
from dataclasses import dataclass, asdict
from deepdiff import DeepHash


class Metadata(NamedTuple):
    n: int
    request_id: str


class PromptNamespace(NamedTuple):
    prompt: str
    namespace: str


@dataclass(frozen=True)
class Request:
    args: Any
    kwargs: Any
    n: int
    request_id: str
    namespace: str

    def __getattr__(self, name: str) -> None:
        return None

    def hash(self) -> int:
        """
        Defines the parameters that uniquely identify a request in the cache.
        """
        params = asdict(self)
        del params["request_id"]
        del params["namespace"]
        del params["n"]
        return DeepHash(params)[params]


class Batch(NamedTuple):
    requests: List[Request]


class Response(NamedTuple):
    data: List[Any]
    cached: Optional[List[bool]] = None
    duplicated: Optional[List[bool]] = None


@runtime_checkable
class SingleRequestModel(Protocol):
    """Model that processes one request at a time, can return multiple results."""

    async def request(self, request: Request) -> Response: ...


@runtime_checkable
class BatchRequestModel(Protocol):
    """Model that processes multiple requests at once, returns multiple results per request."""

    async def batch_request(self, batch: Batch) -> List[Response]: ...



import pytest
import asyncio
from typing import List, Any
from src.cachesaver.typedefs import Request, Batch, BatchRequestModel, Response
from src.cachesaver.batching import AsyncBatcher


class MockBatchModel:
    def __init__(self):
        self.calls: List[Batch] = []
        self.response_delay: float = 0.0

    async def batch_request(self, batch: Batch) -> List[List[Any]]:
        self.calls.append(batch)
        if self.response_delay > 0:
            await asyncio.sleep(self.response_delay)
        return [Response(data=["response"]) for req in batch.requests]


class TestAsyncBatcher:
    @pytest.mark.asyncio
    async def test_batching_mechanism(self):
        """Test that requests are properly batched according to batch_size"""
        mock_model = MockBatchModel()
        batch_size = 2

        async with AsyncBatcher(model=mock_model, batch_size=batch_size) as batcher:
            # Create 5 requests - should result in 3 batches (2, 2, 1)
            requests = [
                Request(args=(), kwargs={"messages": f"test{i}"}, n=1,
                        request_id=str(i), namespace="test")
                for i in range(4)
            ]

            # Send requests concurrently
            tasks = [batcher.request(req) for req in requests]
            await asyncio.gather(*tasks)

            # Verify batching
            assert len(mock_model.calls) == 2
            assert len(mock_model.calls[0].requests) == batch_size
            assert len(mock_model.calls[1].requests) == batch_size

    @pytest.mark.asyncio
    async def test_timeout_mechanism(self):
        """Test that incomplete batches are processed after timeout"""
        mock_model = MockBatchModel()
        batch_size = 3
        timeout = 0.5

        async with AsyncBatcher(model=mock_model, batch_size=batch_size,
                                timeout=timeout) as batcher:
            # Send 2 requests (not enough to fill a batch)
            requests = [
                Request(args=(), kwargs={"messages": f"test{i}"}, n=1,
                        request_id=str(i), namespace="test")
                for i in range(2)
            ]

            start_time = asyncio.get_event_loop().time()
            tasks = [batcher.request(req) for req in requests]
            await asyncio.gather(*tasks)
            elapsed_time = asyncio.get_event_loop().time() - start_time

            # Verify timeout behavior
            assert elapsed_time >= timeout
            assert len(mock_model.calls) == 1
            assert len(mock_model.calls[0].requests) == 2

    @pytest.mark.asyncio
    async def test_multiple_samples_per_request(self):
        """Test handling of n > 1"""
        mock_model = MockBatchModel()
        batch_size = 2

        async with AsyncBatcher(model=mock_model, batch_size=batch_size) as batcher:
            # Create a request asking for 3 samples
            request = Request(args=(), kwargs={"messages": "test"}, n=4,
                              request_id="1", namespace="test")

            results = await batcher.request(request)

            # Verify results and batching
            assert len(results.data) == 4
            assert len(mock_model.calls) == 2  # Should be split into 2 batches
            assert len(mock_model.calls[0].requests) == 2
            assert len(mock_model.calls[1].requests) == 2

    @pytest.mark.asyncio
    async def test_concurrent_requests(self):
        """Test handling multiple concurrent requests"""
        mock_model = MockBatchModel()
        mock_model.response_delay = 0.1  # Add small delay to simulate processing
        batch_size = 2

        async with AsyncBatcher(model=mock_model, batch_size=batch_size) as batcher:
            # Create 4 concurrent requests
            requests = [
                Request(args=(), kwargs={"messages": f"test{i}"}, n=1,
                        request_id=str(i), namespace="test")
                for i in range(4)
            ]

            # Send them all at once
            tasks = [batcher.request(req) for req in requests]
            await asyncio.gather(*tasks)

            # Verify batching under concurrent load
            assert len(mock_model.calls) == 2
            assert all(len(batch.requests) == batch_size
                       for batch in mock_model.calls)

    @pytest.mark.asyncio
    async def test_cleanup_on_context_exit(self):
        """Test that worker task is properly cleaned up when context exits"""
        mock_model = MockBatchModel()
        batch_size = 2

        async with AsyncBatcher(model=mock_model, batch_size=batch_size, timeout=0.1) as batcher:
            # Verify worker task is running
            assert not batcher.worker_task.done()

            # Do some work
            request = Request(args=(), kwargs={"messages": "test"}, n=1,
                              request_id="1", namespace="test")
            await batcher.request(request)

        # Verify worker task is cleaned up
        assert batcher.worker_task.done()

    @pytest.mark.asyncio
    async def test_allow_batch_overflow_single_large_request(self):
        """Test overflow with a single request larger than batch size."""
        mock_model = MockBatchModel()
        batch_size = 5
        timeout = 0.1  # Short timeout

        async with AsyncBatcher(
            model=mock_model,
            batch_size=batch_size,
            timeout=timeout,
            allow_batch_overflow=True
        ) as batcher:
            # Single request larger than batch size
            request1 = Request(args=(), kwargs={"messages": "large"}, n=7, request_id="r1", namespace="test")
            result1 = await batcher.request(request1)

            # Verify: Should process as one batch despite size
            assert len(result1.data) == 7
            assert len(mock_model.calls) == 1
            assert len(mock_model.calls[0].requests) == 7

    @pytest.mark.asyncio
    async def test_allow_batch_overflow_fills_existing(self):
        """Test overflow where a request fills and exceeds existing items."""
        mock_model = MockBatchModel()
        batch_size = 5
        timeout = 3  # Short timeout

        async with AsyncBatcher(
            model=mock_model,
            batch_size=batch_size,
            timeout=timeout,
            allow_batch_overflow=True
        ) as batcher:
            # Send 3 initial requests (batch size 5)
            initial_requests = [
                Request(args=(), kwargs={"messages": f"init{i}"}, n=1, request_id=f"i{i}", namespace="test")
                for i in range(3)
            ]
            replies = [batcher.request(req) for req in initial_requests]

            # Create extra request that leads to overflow
            overflow_request = Request(args=(), kwargs={"messages": "overflow"}, n=7, request_id="of1", namespace="test")

            overflow_reply = batcher.request(overflow_request)

            # Await the results
            replies.append(overflow_reply)
            all_results = await asyncio.gather(*replies)

            # Extract results for verification
            initial_results = all_results[:3]
            overflow_result = all_results[3]

            # Verify: Should still be one batch with all 10 requests
            assert len([r.data[0] for r in initial_results]) == 3
            assert len(overflow_result.data) == 7
            assert len(mock_model.calls) == 1
            assert len(mock_model.calls[0].requests) == 10 # 3 initial + 7 overflow




import pytest
from typing import List, Any
from src.cachesaver.typedefs import Request, Response, Batch, BatchRequestModel
from src.cachesaver.deduplicator import AsyncDeduplicator
from collections import defaultdict


class MockBatchModel:
    def __init__(self):
        self.calls: List[Batch] = []
        self.responses: List[List[Any]] = []

    async def batch_request(self, batch: Batch) -> List[Response]:
        self.calls.append(batch)
        # Generate responses based on n
        responses = []
        for request in batch.requests:
            responses.append(
                Response(data=[f"response_{i}" for i in range(request.n)]))
        return responses


class DeterministicMockModel:
    def __init__(self):
        self.counters = defaultdict(int)

    async def batch_request(self, batch: Batch) -> List[Response]:
        responses = []
        for request in batch.requests:
            request_responses = []
            prompt = request.kwargs.get('messages', '')
            key = (request.namespace, prompt)
            for _ in range(request.n):
                response = f"{prompt}_{request.namespace}_{self.counters[key]}"
                self.counters[key] += 1
                request_responses.append(response)
            responses.append(Response(data=request_responses))
        return responses


class TestAsyncDeduplicator:
    @pytest.mark.asyncio
    async def test_deduplication_same_namespace(self):
        """Test that identical prompts in same namespace are deduplicated"""
        mock_model = MockBatchModel()
        deduplicator = AsyncDeduplicator(model=mock_model)

        # Create requests with same prompt and namespace
        batch = Batch(requests=[
            Request(args=(), kwargs={"messages": "test"}, n=2,
                    request_id="1", namespace="test"),
            Request(args=(), kwargs={"messages": "test"}, n=3,
                    request_id="2", namespace="test"),
            Request(args=(), kwargs={"messages": "test"}, n=1,
                    request_id="3", namespace="test")
        ])

        results = await deduplicator.batch_request(batch)

        # Verify deduplication
        assert len(mock_model.calls) == 1  # Should be merged into single call
        merged_request = mock_model.calls[0].requests[0]
        assert merged_request.n == 6  # Sum of all n
        assert merged_request.kwargs.get("messages") == "test"
        assert merged_request.namespace == "test"

        # Verify results were split correctly
        assert len(results) == 3
        assert len(results[0].data) == 2  # First request wanted 2
        assert len(results[1].data) == 3  # Second request wanted 3
        assert len(results[2].data) == 1  # Third request wanted 1

    @pytest.mark.asyncio
    async def test_no_deduplication_different_namespaces(self):
        """Test that prompts in different namespaces are not deduplicated"""
        mock_model = MockBatchModel()
        deduplicator = AsyncDeduplicator(model=mock_model)

        # Create requests with same prompt but different namespaces
        batch = Batch(requests=[
            Request(args=(), kwargs={"messages": "test"}, n=2,
                    request_id="1", namespace="ns1"),
            Request(args=(), kwargs={"messages": "test"}, n=2,
                    request_id="2", namespace="ns2")
        ])

        results = await deduplicator.batch_request(batch)

        # Verify no deduplication occurred
        assert len(mock_model.calls) == 1  # Still one batch call
        # But with two separate requests
        assert len(mock_model.calls[0].requests) == 2

        # Verify each request maintained its namespace
        requests = mock_model.calls[0].requests
        assert requests[0].namespace != requests[1].namespace

        # Verify results
        assert len(results) == 2
        assert all(len(result.data) == 2 for result in results)

    @pytest.mark.asyncio
    async def test_mixed_deduplication(self):
        """Test handling of mixed duplicates and unique requests"""
        mock_model = MockBatchModel()
        deduplicator = AsyncDeduplicator(model=mock_model)

        batch = Batch(requests=[
            Request(args=(), kwargs={"messages": "test1"}, n=2,
                    request_id="1", namespace="ns1"),
            Request(args=(), kwargs={"messages": "test1"}, n=3,
                    request_id="2", namespace="ns1"),
            Request(args=(), kwargs={"messages": "test2"}, n=1,
                    request_id="3", namespace="ns1"),
            Request(args=(), kwargs={"messages": "test1"}, n=2,
                    request_id="4", namespace="ns2")
        ])

        results = await deduplicator.batch_request(batch)

        # Verify partial deduplication
        assert len(mock_model.calls) == 1
        requests = mock_model.calls[0].requests
        assert len(requests) == 3  # Should have 3 requests after deduplication

        # Verify results
        assert len(results) == 4  # Still 4 original results
        assert len(results[0].data) == 2
        assert len(results[1].data) == 3
        assert len(results[2].data) == 1
        assert len(results[3].data) == 2

    @pytest.mark.asyncio
    async def test_response_order_preservation(self):
        """Test that deduplication doesn't change the order or content of responses"""
        mock_model = DeterministicMockModel()
        deduplicator = AsyncDeduplicator(model=mock_model)

        # Create a batch with various patterns of duplicates
        batch = Batch(requests=[
            Request(args=(), kwargs={"messages": "p1"}, n=2,
                    request_id="1", namespace="ns1"),
            Request(args=(), kwargs={"messages": "p1"}, n=1, request_id="2",
                    namespace="ns1"),  # Duplicate
            Request(args=(), kwargs={"messages": "p2"}, n=1,
                    request_id="3", namespace="ns1"),
            Request(args=(), kwargs={"messages": "p1"}, n=2, request_id="4",
                    namespace="ns2"),  # Same prompt, different ns
            Request(args=(), kwargs={"messages": "p2"}, n=1, request_id="5",
                    namespace="ns1"),  # Duplicate of p2
        ])

        # Get responses with deduplication
        mock_model.counter = 0  # Reset counter
        results_with_dedup = await deduplicator.batch_request(batch)

        # Get responses without deduplication (directly from model)
        mock_model = DeterministicMockModel()
        results_without_dedup = await mock_model.batch_request(batch)

        # Compare results
        assert len(results_with_dedup) == len(results_without_dedup)
        for with_dedup, without_dedup in zip(results_with_dedup, results_without_dedup):
            for wd, woutd in zip(with_dedup.data, without_dedup.data):
                assert wd == woutd, (
                    f"Mismatch in results: "
                    f"with dedup: {wd}, "
                    f"without dedup: {woutd}"
                )



import pytest
import asyncio
import tempfile
from collections import defaultdict
from typing import List, Any
from diskcache import Cache

from src.cachesaver.typedefs import Request, Response, Batch, SingleRequestModel
from src.cachesaver.pipelines import OnlineAPI


class CountingMockModel(SingleRequestModel):
    """Mock model that tracks individual requests and returns incrementing responses."""

    def __init__(self):
        self.counter = 0
        self.calls: List[Request] = []  # Track individual requests

    async def request(self, request: Request) -> List[Any]:
        self.calls.append(request)
        responses = []
        prompt = request.kwargs.get('messages', '')
        for _ in range(request.n):
            responses.append(f"{prompt}_{self.counter}")
            self.counter += 1
        return Response(data=responses)


class OrderTrackingModel(SingleRequestModel):
    """Mock model that tracks request order and maintains per-prompt counters."""

    def __init__(self):
        self.prompt_counters = defaultdict(int)
        self.calls: List[Request] = []

    async def request(self, request: Request) -> List[Any]:
        self.calls.append(request)
        responses = []
        prompt = request.kwargs.get('messages', '')
        for _ in range(request.n):
            counter = self.prompt_counters[prompt]
            responses.append(f"{prompt}_count{counter}")
            self.prompt_counters[prompt] += 1
        return Response(data=responses)


class TestOnlineAPI:
    TEST_TIMEOUT = 0.1  # Short timeout for tests

    @pytest.fixture
    def cache(self):
        """Provide a temporary cache for each test."""
        with tempfile.TemporaryDirectory() as tmpdir:
            with Cache(tmpdir) as cache:
                yield cache

    @pytest.mark.asyncio
    async def test_reuse_across_namespaces(self, cache):
        """Test that identical prompts are deduplicated and cached."""
        model = CountingMockModel()

        async with OnlineAPI(
            model=model,
            cache=cache,
            batch_size=3,
            timeout=self.TEST_TIMEOUT
        ) as api:
            # Create requests with identical prompts, but different namespaces
            requests = [
                Request(args=(), kwargs={"messages": "test_prompt"}, n=1,
                        request_id="1", namespace="ns1"),
                Request(args=(), kwargs={"messages": "test_prompt"}, n=1,
                        request_id="2", namespace="ns2")
            ]

            # First batch of requests
            results1 = await asyncio.gather(*[api.request(req) for req in requests])

            # Same prompt in new namespace
            request3 = Request(args=(), kwargs={"messages": "test_prompt"}, n=1,
                               request_id="3", namespace="ns3")
            result3 = await api.request(request3)

            # Verify deduplication and caching
            assert len(model.calls) == 1  # Single request to model
            assert model.calls[0].n == 1  # One sample reused

    @pytest.mark.asyncio
    async def test_deduplication_within_namespace(self, cache):
        """Test that requests in same namespace get fresh responses."""
        model = CountingMockModel()

        async with OnlineAPI(
            model=model,
            cache=cache,
            batch_size=2,
            timeout=self.TEST_TIMEOUT
        ) as api:
            # Two identical requests in same namespace
            requests = [
                Request(args=(), kwargs={"messages": "test"}, n=1,
                        request_id="1", namespace="same_ns"),
                Request(args=(), kwargs={"messages": "test"}, n=1,
                        request_id="2", namespace="same_ns")
            ]

            results = await asyncio.gather(*[api.request(req) for req in requests])

            # Verify separate responses in same namespace
            assert len(model.calls) == 1  # One request to model
            assert results[0] != results[1]  # Different responses
            assert model.calls[0].namespace == "same_ns"

    @pytest.mark.asyncio
    async def test_batching_behavior(self, cache):
        """Test that requests are collected into batches before processing."""
        model = CountingMockModel()

        async with OnlineAPI(
            model=model,
            cache=cache,
            batch_size=2,
            timeout=self.TEST_TIMEOUT
        ) as api:
            # Three unique requests
            requests = [
                Request(args=(), kwargs={"messages": f"unique_{i}"}, n=1,
                        request_id=str(i), namespace="ns1")
                for i in range(3)
            ]

            await asyncio.gather(*[api.request(req) for req in requests])

            # Verify batching before deduplication
            assert len(model.calls) == 3  # One call per unique prompt
            # First two should be processed together
            assert model.calls[0].kwargs.get('messages', '').startswith("unique_0")
            assert model.calls[1].kwargs.get('messages', '').startswith("unique_1")
            # Last one in separate batch
            assert model.calls[2].kwargs.get('messages', '').startswith("unique_2")

    @pytest.mark.asyncio
    async def test_mixed_scenario(self, cache):
        """Test combination of batching, deduplication, and caching."""
        model = CountingMockModel()

        async with OnlineAPI(
            model=model,
            cache=cache,
            batch_size=3,
            timeout=self.TEST_TIMEOUT
        ) as api:
            # Mix of duplicate and unique requests
            requests = [
                Request(args=(), kwargs={"messages": "shared"}, n=1,
                        request_id="1", namespace="ns1"),
                Request(args=(), kwargs={"messages": "shared"}, n=1,
                        request_id="2", namespace="ns1"),  # Should be deduplicated
                Request(args=(), kwargs={"messages": "unique"}, n=1,
                        request_id="3", namespace="ns2")
            ]

            results1 = await asyncio.gather(*[api.request(req) for req in requests])

            # Additional request that should use cache
            request4 = Request(args=(), kwargs={"messages": "shared"}, n=1,
                               request_id="4", namespace="ns3")
            result4 = await api.request(request4)

            # Verify optimizations
            # One for deduplicated "shared", one for "unique"
            assert len(model.calls) == 2
            assert model.calls[0].kwargs.get('messages') == "shared"
            assert model.calls[0].n == 2  # Merged requests
            assert model.calls[1].kwargs.get('messages') == "unique"


class TestOnlineAPIOrdering:
    TEST_TIMEOUT = 0.1

    @pytest.fixture
    def cache(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            with Cache(tmpdir) as cache:
                yield cache

    @pytest.mark.asyncio
    async def test_ordered_batch_processing(self, cache):
        """Test that responses maintain order based on request_ids."""
        model = OrderTrackingModel()

        async with OnlineAPI(
            model=model,
            cache=cache,
            batch_size=3,
            timeout=self.TEST_TIMEOUT
        ) as api:
            # Create requests with out-of-order IDs
            requests = [
                Request(args=(), kwargs={"messages": "p1"}, n=1,
                        request_id="3", namespace="ns1"),
                Request(args=(), kwargs={"messages": "p1"}, n=1,
                        request_id="1", namespace="ns1"),
                Request(args=(), kwargs={"messages": "p1"}, n=1,
                        request_id="2", namespace="ns1")
            ]

            # Send requests concurrently
            results = await asyncio.gather(*[api.request(req) for req in requests])

            # Results should maintain original order (3,1,2) but responses should be ordered
            assert results[0].data[0] == "p1_count2"  # request_id "3"
            assert results[1].data[0] == "p1_count0"  # request_id "1"
            assert results[2].data[0] == "p1_count1"  # request_id "2"

    @pytest.mark.asyncio
    async def test_ordering_with_cache_hits(self, cache):
        """Test ordering when some responses come from cache."""
        model = OrderTrackingModel()

        async with OnlineAPI(
            model=model,
            cache=cache,
            batch_size=2,
            timeout=self.TEST_TIMEOUT
        ) as api:
            # First request to populate cache
            await api.request(
                Request(args=(), kwargs={"messages": "cached"}, n=1,
                        request_id="0", namespace="ns1")
            )

            # Mix of cached and new requests with out-of-order IDs
            requests = [
                Request(args=(), kwargs={"messages": "new"}, n=1,
                        request_id="3", namespace="ns1"),
                Request(args=(), kwargs={"messages": "cached"}, n=1,
                        request_id="1", namespace="ns2"),
                Request(args=(), kwargs={"messages": "new"}, n=1,
                        request_id="2", namespace="ns1")
            ]

            results = await asyncio.gather(*[api.request(req) for req in requests])

            # Verify ordering is maintained even with cache hits
            assert results[0].data[0].startswith("new_count")      # request_id "3"
            # request_id "1" (from cache)
            assert results[1].data[0].startswith("cached_count")
            assert results[2].data[0].startswith("new_count")      # request_id "2"

    @pytest.mark.asyncio
    async def test_ordering_with_deduplication(self, cache):
        """Test ordering when requests are deduplicated."""
        model = OrderTrackingModel()

        async with OnlineAPI(
            model=model,
            cache=cache,
            batch_size=3,
            timeout=self.TEST_TIMEOUT
        ) as api:
            # Create requests with duplicate prompts but out-of-order IDs
            requests = [
                Request(args=(), kwargs={"messages": "dup"}, n=1,
                        request_id="3", namespace="ns1"),
                Request(args=(), kwargs={"messages": "dup"}, n=1,
                        request_id="1", namespace="ns2"),
                Request(args=(), kwargs={"messages": "unique"}, n=1,
                        request_id="2", namespace="ns1")
            ]

            results = await asyncio.gather(*[api.request(req) for req in requests])

            # Verify responses maintain order despite deduplication
            assert results[0].data[0].startswith("dup_count")      # request_id "3"
            assert results[1].data[0].startswith("dup_count")      # request_id "1"
            assert results[2].data[0].startswith("unique_count")   # request_id "2"

            # Verify deduplication occurred
            assert len([c for c in model.calls if c.kwargs.get('messages') == "dup"]) == 1

    @pytest.mark.asyncio
    async def test_batch_interface_ordering(self, cache):
        """Test ordering when using batch_request interface directly."""
        model = OrderTrackingModel()

        async with OnlineAPI(
            model=model,
            cache=cache,
            batch_size=4,
            timeout=self.TEST_TIMEOUT
        ) as api:
            batch = Batch(requests=[
                Request(args=(), kwargs={"messages": "p1"}, n=1,
                        request_id="3", namespace="ns1"),
                Request(args=(), kwargs={"messages": "p2"}, n=1,
                        request_id="1", namespace="ns1"),
                Request(args=(), kwargs={"messages": "p1"}, n=1,
                        request_id="4", namespace="ns2"),
                Request(args=(), kwargs={"messages": "p2"}, n=1,
                        request_id="2", namespace="ns2")
            ])

            results = await api.batch_request(batch)

            # Verify responses maintain original batch order but are internally ordered
            assert results[0].data[0].startswith("p1_count")   # request_id "3"
            assert results[1].data[0].startswith("p2_count")   # request_id "1"
            assert results[2].data[0].startswith("p1_count")   # request_id "4"
            assert results[3].data[0].startswith("p2_count")   # request_id "2"


class TestOnlineAPIBatching:
    TEST_TIMEOUT = 0.1  # Short timeout for tests

    @pytest.fixture
    def cache(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            with Cache(tmpdir) as cache:
                yield cache
    
    @pytest.mark.asyncio
    async def test_allow_batch_overflow_single_large_request(self, cache):
        """Test overflow with a single request larger than batch size."""
        model = CountingMockModel()
        batch_size = 5

        async with OnlineAPI(
            model=model,
            cache=cache,
            batch_size=batch_size,
            timeout=self.TEST_TIMEOUT,
            allow_batch_overflow=True
        ) as api:
            # Single request larger than batch size
            request1 = Request(args=(), kwargs={"messages": "large"}, n=7, request_id="r1", namespace="test")
            result1 = await api.request(request1)

            # Verify: Should process as one batch despite size
            assert len(result1.data) == 7
            assert len(model.calls) == 1
            assert model.calls[0].n == 7
    
    @pytest.mark.asyncio
    async def test_allow_batch_overflow_fills_existing(self, cache):
        """Test overflow where a request fills and exceeds existing items."""
        model = CountingMockModel()
        batch_size = 5

        async with OnlineAPI(
            model=model,
            cache=cache,
            batch_size=batch_size,
            timeout=3,
            allow_batch_overflow=True
        ) as api:
            # Send 3 initial requests (batch size 5)
            initial_requests = [
                Request(args=(), kwargs={"messages": f"init{i}"}, n=1, request_id=f"i{i}", namespace="test")
                for i in range(3)
            ]
            replies = [api.request(req) for req in initial_requests]

            # Create extra request that leads to overflow
            overflow_request = Request(args=(), kwargs={"messages": "overflow"}, n=7, request_id="of1", namespace="test")
            overflow_reply = api.request(overflow_request)

            # Await the results
            replies.append(overflow_reply)
            all_results = await asyncio.gather(*replies)

            # Extract results for verification
            initial_results = all_results[:3]
            overflow_result = all_results[3]

            # Verify: Should still be one batch with all 10 requests
            assert len([r.data[0] for r in initial_results]) == 3
            assert len(overflow_result.data) == 7
            assert len(model.calls) == 4
            assert model.calls[-1].n == 7 



from ..pipelines import OnlineAPI
from diskcache import Cache
from copy import deepcopy
from ..typedefs import BatchRequestModel, Batch, Request, Metadata, SingleRequestModel, Response

class Create:
    def __init__(self, api, namespace):
        self.api = api
        self.global_request_counter = 0
        self.namespace = namespace

    async def __call__(self, *args, **kwargs):
        self.global_request_counter += 1
        if "n" in kwargs:
            n = kwargs.pop("n")
        else:
            n = 1
        flattened_responses = await self.api.request(Request(
            args=args,
            kwargs=kwargs,
            n=n,
            namespace=self.namespace,
            request_id=str(self.global_request_counter)
        ))
        choices = [resp.choices[0] for resp in flattened_responses.data]
        one_response = deepcopy(flattened_responses.data[0])
        one_response.choices = choices
        return one_response


class Completions:
    def __init__(self, api, namespace):
        self.create = Create(api, namespace)


class Chat:
    def __init__(self,  api, namespace):
        self.completions = Completions( api, namespace)


class AsyncWrapper:
    def __init__(self,
                 model:SingleRequestModel,
                 namespace="default",
                 cachedir="./cache",
                 batch_size=1):
        self.namespace = namespace
        self.cachedir = cachedir
        self.batch_size = batch_size
        self.cache = Cache(self.cachedir, timeout=-1)
        self.model = model
        self.api = OnlineAPI(model=self.model,
                             cache=self.cache, batch_size=self.batch_size)
        self.chat = Chat(self.api, self.namespace)



from typing import Optional
from diskcache import Cache

from .typedefs import SingleRequestModel, BatchRequestModel
from .batching import AsyncBatcher
from .deduplicator import AsyncDeduplicator
from .caching import AsyncCacher
from .reordering import RequestReorderer
from .resources import AsyncResource


class LocalAPI(AsyncResource, SingleRequestModel, BatchRequestModel):
    """
    Pipeline optimized for local model inference.

    Flow: Cache -> Batcher -> Model
    Optimizes for full batch utilization while avoiding unnecessary model calls.
    """

    def __init__(
        self,
        model: BatchRequestModel,
        cache: Cache,
        batch_size: int,
        timeout: int = 30,
        allow_batch_overflow: bool = False
    ):
        # Create pipeline from inside out
        self.batcher = AsyncBatcher(
            model=model,
            batch_size=batch_size,
            timeout=timeout,
            name="local_batcher",
            allow_batch_overflow=allow_batch_overflow
        )
        self.cacher = AsyncCacher(model=self.batcher, cache=cache)

    async def request(self, request):
        """Single request API, delegates to cache."""
        return await self.cacher.request(request)

    async def batch_request(self, batch):
        """Batch request API, delegates to cache."""
        return await self.cacher.batch_request(batch)

    async def cleanup(self):
        """Cleanup all pipeline components."""
        await self.batcher.cleanup()
        if hasattr(self.cacher, 'cleanup'):
            await self.cacher.cleanup()


class OnlineAPI(AsyncResource, SingleRequestModel, BatchRequestModel):
    """
    Pipeline optimized for cloud API cost reduction.

    Flow: Batcher -> Reorderer -> Deduplicator -> Cache -> Model
    """

    def __init__(
        self,
        model: BatchRequestModel,
        cache: Cache,
        batch_size: int,
        timeout: int = 30,
        allow_batch_overflow: bool = False,
        correctness: bool = False # Happy to remove this if not needed
    ):
        # Create pipeline from inside out
        self.cached_model = AsyncCacher(model=model, cache=cache)
        self.deduplicator = AsyncDeduplicator(model=self.cached_model, correctness=correctness)
        self.reorderer = RequestReorderer(model=self.deduplicator)
        self.batcher = AsyncBatcher(
            model=self.reorderer,
            batch_size=batch_size,
            timeout=timeout,
            name="online_batcher",
            allow_batch_overflow=allow_batch_overflow
        )

    async def request(self, request):
        """Single request API, delegates to batcher."""
        return await self.batcher.request(request)

    async def batch_request(self, batch):
        """Batch request API, delegates to batcher."""
        return await self.batcher.batch_request(batch)

    async def cleanup(self):
        """Cleanup all pipeline components."""
        await self.batcher.cleanup()
        if hasattr(self.cached_model, 'cleanup'):
            await self.cached_model.cleanup()


class OrderedLocalAPI(AsyncResource, SingleRequestModel, BatchRequestModel):
    """
    Pipeline optimized for local model inference with guaranteed ordering.

    Flow: Collect Batch -> Reorder -> Cache -> Hardware Batch -> Model

    This pipeline ensures:
    1. Responses are ordered by request_id (reproducible results)
    2. Cache is checked for existing responses
    3. Remaining requests are batched for efficient hardware utilization
    """

    def __init__(
        self,
        model: BatchRequestModel,
        cache: Cache,
        collection_batch_size: int,
        hardware_batch_size: int,
        timeout: int = 30,
        allow_batch_overflow: bool = False
    ):
        # Create pipeline from inside out (model -> user)
        self.hardware_batcher = AsyncBatcher(
            model=model,
            batch_size=hardware_batch_size,
            timeout=timeout,
            name="hardware_batcher"
        )
        self.cacher = AsyncCacher(
            model=self.hardware_batcher,
            cache=cache
        )
        self.deduplicator = AsyncDeduplicator(
            model=self.cacher
        )
        self.reorderer = RequestReorderer(
            model=self.deduplicator
        )
        self.collection_batcher = AsyncBatcher(
            model=self.reorderer,
            batch_size=collection_batch_size,
            timeout=timeout,
            name="collection_batcher",
            allow_batch_overflow=allow_batch_overflow
        )

    async def request(self, request):
        """Single request API, delegates to collection batcher."""
        return await self.collection_batcher.request(request)

    async def batch_request(self, batch):
        """Batch request API, delegates to collection batcher."""
        return await self.collection_batcher.batch_request(batch)

    async def cleanup(self):
        """Cleanup all pipeline components."""
        await self.collection_batcher.cleanup()
        await self.hardware_batcher.cleanup()
        if hasattr(self.cacher, 'cleanup'):
            await self.cacher.cleanup()



from openai import AsyncClient
from openai import ChatCompletion
from .pipelines import OnlineAPI
from diskcache import Cache
from copy import deepcopy
from .typedefs import BatchRequestModel, Batch, Request, Metadata, SingleRequestModel, Response


class OpenAIModel(SingleRequestModel):
    def __init__(self, aclient: AsyncClient):
        self.aclient = aclient

    async def request(self, request: Request):
        args = request.args
        kwargs = request.kwargs

        response = await self.aclient.chat.completions.create(
            *args, **kwargs, n=request.n
        )

        # ToDo: this is difficult, how do we deal with the case n>1
        # Cachesaver then expects a list of replies
        flattened_responses = []
        for choice in response.choices:
            response_copy = deepcopy(response)
            response_copy.choices = [choice]
            flattened_responses.append(response_copy)

        return Response(
            data=flattened_responses)


class Create:
    def __init__(self, aclient, api, namespace):
        self.aclient = aclient
        self.api = api
        self.global_request_counter = 0
        self.namespace = namespace

    async def __call__(self, *args, **kwargs):
        self.global_request_counter += 1
        if "n" in kwargs:
            n = kwargs.pop("n")
        else:
            n = 1
        flattened_responses = await self.api.request(Request(
            args=args,
            kwargs=kwargs,
            n=n,
            namespace=self.namespace,
            request_id=str(self.global_request_counter)
        ))
        choices = [resp.choices[0] for resp in flattened_responses.data]
        one_response = deepcopy(flattened_responses.data[0])
        one_response.choices = choices
        return one_response


class Completions:
    def __init__(self, aclient, api, namespace):
        self.create = Create(aclient, api, namespace)


class Chat:
    def __init__(self, aclient, api, namespace):
        self.completions = Completions(aclient, api, namespace)


class AsyncWrapper:
    def __init__(self,
                 namespace="default",
                 cachedir="./cache",
                 batch_size=1):
        self.namespace = namespace
        self.cachedir = cachedir
        self.batch_size = batch_size
        self.cache = Cache("./cache", timeout=-1)
        self.aclient = AsyncClient()
        self.model = OpenAIModel(self.aclient)
        self.api = OnlineAPI(model=self.model,
                             cache=self.cache, batch_size=self.batch_size)
        self.chat = Chat(self.aclient, self.api, self.namespace)



import pytest
import tempfile
from typing import List, Any
from diskcache import Cache
import asyncio
from src.cachesaver.typedefs import Request, Response, Batch, SingleRequestModel
from src.cachesaver.caching import AsyncCacher


class CountingMockModel(SingleRequestModel):
    """Mock model that returns incrementing responses to track call order."""

    def __init__(self):
        self.counter = 0
        self.calls = []

    async def request(self, request: Request) -> List[Any]:
        self.calls.append(request)
        responses = []
        for _ in range(request.n):
            responses.append(f"response_{self.counter}")
            self.counter += 1
        return Response(data=responses)


class TestAsyncCacher:
    @pytest.fixture
    def cache(self):
        """Provide a temporary cache for each test."""
        with tempfile.TemporaryDirectory() as tmpdir:
            with Cache(tmpdir) as cache:
                yield cache

    @pytest.mark.asyncio
    async def test_basic_caching(self, cache):
        """Test that responses are cached and reused."""
        model = CountingMockModel()
        cacher = AsyncCacher(model=model, cache=cache)

        # First request should hit the model
        request = Request(args=(), kwargs={"messages": "test"}, n=2,
                          request_id="1", namespace="ns1")
        result1 = await cacher.request(request)
        assert result1.data == ["response_0", "response_1"]
        assert len(model.calls) == 1

        # Same request in different namespace should reuse cache
        request = Request(args=(), kwargs={"messages": "test"}, n=2,
                          request_id="2", namespace="ns2")
        result2 = await cacher.request(request)
        assert result2.data == ["response_0", "response_1"]
        assert len(model.calls) == 1  # No new model calls

    @pytest.mark.asyncio
    async def test_namespace_isolation(self, cache):
        """Test that requests within same namespace get fresh responses."""
        model = CountingMockModel()
        cacher = AsyncCacher(model=model, cache=cache)

        # First request
        request1 = Request(args=(), kwargs={"messages": "test"}, n=1,
                           request_id="1", namespace="same_ns")
        result1 = await cacher.request(request1)
        assert result1.data == ["response_0"]

        # Same prompt, same namespace should get fresh response
        request2 = Request(args=(), kwargs={"messages": "test"}, n=1,
                           request_id="2", namespace="same_ns")
        result2 = await cacher.request(request2)
        assert result2.data == ["response_1"]  # New response
        assert len(model.calls) == 2  # Required new model call

    @pytest.mark.asyncio
    async def test_batch_request_caching(self, cache):
        """Test that batch requests properly use cache."""
        model = CountingMockModel()
        cacher = AsyncCacher(model=model, cache=cache)

        batch = Batch(requests=[
            Request(args=(), kwargs={"messages": "p1"}, n=2,
                    request_id="1", namespace="ns1"),
            Request(args=(), kwargs={"messages": "p1"}, n=1, request_id="2",
                    namespace="ns2"),  # Should reuse
            Request(args=(), kwargs={"messages": "p2"}, n=1,
                    request_id="3", namespace="ns1")
        ])

        results = await cacher.batch_request(batch)
        assert len(results) == 3
        assert results[0].data == ["response_0", "response_1"]  # First p1 request
        assert results[1].data == ["response_0"]  # Reused from cache
        assert results[2].data == ["response_2"]  # New p2 request

        assert len(model.calls) == 2  # Only needed calls for p1/ns1 and p2/ns1

    @pytest.mark.asyncio
    async def test_partial_cache_reuse(self, cache):
        """Test requesting more responses than cached."""
        model = CountingMockModel()
        cacher = AsyncCacher(model=model, cache=cache)

        # Initial request caches 2 responses
        request1 = Request(args=(), kwargs={"messages": "test"}, n=2,
                           request_id="1", namespace="ns1")
        result1 = await cacher.request(request1)
        assert result1.data == ["response_0", "response_1"]

        # Request more responses in different namespace
        request2 = Request(args=(), kwargs={"messages": "test"}, n=3,
                           request_id="2", namespace="ns2")
        result2 = await cacher.request(request2)
        # Should reuse first 2 responses and get 1 new
        assert result2.data == ["response_0", "response_1", "response_2"]
        assert len(model.calls) == 2

    @pytest.mark.asyncio
    async def test_concurrent_access(self, cache):
        """Test that concurrent requests are properly synchronized."""
        model = CountingMockModel()
        cacher = AsyncCacher(model=model, cache=cache)

        # Create multiple concurrent requests for same prompt
        requests = [
            Request(args=(), kwargs={"messages": "test"}, n=1,
                    request_id=str(i), namespace=f"ns{i}")
            for i in range(3)
        ]

        results = await asyncio.gather(*(cacher.request(r) for r in requests))

        # First response should be used for all requests
        assert all(r.data == ["response_0"] for r in results)
        assert len(model.calls) == 1  # Should only call model once



import logging
from abc import ABC, abstractmethod
from typing import TypeVar, AsyncContextManager

logger = logging.getLogger(__name__)

T = TypeVar('T', bound='AsyncResource')


class AsyncResource(AsyncContextManager[T], ABC):
    """Base class for resources that need async cleanup."""

    async def __aenter__(self: T) -> T:
        """Enter async context."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exit async context and cleanup resources."""
        await self.cleanup()

    @abstractmethod
    async def cleanup(self):
        """Cleanup resources. Must be implemented by subclasses."""
        pass



import pytest
from typing import List, Any
from src.cachesaver.typedefs import SingleRequestModel, BatchRequestModel, Request, Batch, Response


class TestSingleRequestModel:
    class ValidModel:
        async def request(self, request: Request) -> List[Any]:
            return ["response"]

    class InvalidModel:
        async def wrong_method(self, request: Request) -> List[Any]:
            return ["response"]

    def test_protocol_compliance(self):
        valid_model = self.ValidModel()
        invalid_model = self.InvalidModel()
        assert isinstance(valid_model, SingleRequestModel)
        assert not isinstance(invalid_model, SingleRequestModel)

    @pytest.mark.asyncio
    async def test_valid_implementation(self):
        model = self.ValidModel()
        request = Request(args=(), kwargs={"messages": "test"}, n=1,
                          request_id="1", namespace="test")
        result = await model.request(request)
        assert isinstance(result, list)
        assert isinstance(result[0], str)


class TestBatchRequestModel:
    class ValidModel:
        async def batch_request(self, batch: Batch) -> List[List[Any]]:
            return [["response"] for _ in batch.requests]

    class InvalidModel:
        async def different_method_name(self, batch: Batch) -> List[List[Any]]:
            return [["response"] for _ in batch.requests]

    def test_protocol_compliance(self):
        valid_model = self.ValidModel()
        invalid_model = self.InvalidModel()
        assert isinstance(valid_model, BatchRequestModel)
        assert not isinstance(invalid_model, BatchRequestModel)

    @pytest.mark.asyncio
    async def test_valid_implementation(self):
        model = self.ValidModel()
        batch = Batch(requests=[
            Request(args=(), kwargs={"messages": "test1"}, n=1,
                    request_id="1", namespace="test"),
            Request(args=(), kwargs={"messages": "test2"}, n=2,
                    request_id="2", namespace="test")
        ])
        result = await model.batch_request(batch)
        assert isinstance(result, list)
        assert all(isinstance(x, list) for x in result)
        assert all(isinstance(item, str)
                   for sublist in result for item in sublist)


class TestAsyncBatcher:
    class ValidBatchModel:
        async def batch_request(self, batch: Batch) -> List[Response]:
            return [Response(data=["response"]) for _ in batch.requests]

    @pytest.mark.asyncio
    async def test_protocol_compliance(self):
        from src.cachesaver.batching import AsyncBatcher
        async with AsyncBatcher(model=self.ValidBatchModel(), batch_size=2) as batcher:
            assert isinstance(batcher, SingleRequestModel)
            assert isinstance(batcher, BatchRequestModel)

    @pytest.mark.asyncio
    async def test_batch_request(self):
        from src.cachesaver.batching import AsyncBatcher
        async with AsyncBatcher(model=self.ValidBatchModel(), batch_size=2) as batcher:
            batch = Batch(requests=[
                Request(args=(), kwargs={"messages": "test1"}, n=1,
                        request_id="1", namespace="test"),
                Request(args=(), kwargs={"messages": "test2"}, n=3,
                        request_id="2", namespace="test")
            ])
            result = await batcher.batch_request(batch)
            assert isinstance(result, list)
            assert all(isinstance(x, Response) for x in result)



import asyncio
from contextlib import AbstractAsyncContextManager
from typing import Any


class AsyncRoundRobinLimiter(AbstractAsyncContextManager):

    def __init__(self):
        self.queue = asyncio.Queue()
        self.resources = []

    async def __aenter__(self):
        resource = await self.queue.get()
        assert resource.in_use is False, "Resource is already in use"
        resource.in_use = True
        return resource

    def add_resource(self, data: Any = None):
        resource = AsyncRoundRobinResource(self, data)
        self.resources.append(resource)
        self.queue.put_nowait(resource)

    async def __aexit__(self, exc_type, exc, tb):
        return None


class AsyncRoundRobinResource:

    def __init__(self, limiter: AsyncRoundRobinLimiter, data: Any = None):
        self.data = data
        self.in_use = False
        self.limiter = limiter

    def reschedule(self, time_taken: float, amount_used: float):
        self.limiter.queue.put_nowait(self)

    def free(self, time_taken: float, amount_used: float):
        self.reschedule(time_taken, amount_used)
        self.in_use = False



import pytest
import tempfile
from typing import List, Any
from diskcache import Cache
import asyncio

from src.cachesaver.typedefs import Request, Response, Batch, BatchRequestModel
from src.cachesaver.pipelines import LocalAPI


class BatchTrackingModel(BatchRequestModel):
    """Mock model that tracks batch sizes and returns deterministic responses."""

    def __init__(self):
        self.batch_calls: List[Batch] = []
        self.counter = 0

    async def batch_request(self, batch: Batch) -> List[Response]:
        self.batch_calls.append(batch)
        responses = []
        for request in batch.requests:
            request_responses = []
            prompt = request.kwargs.get('messages', '')
            for _ in range(request.n):
                response = f"{prompt}_{request.namespace}_{self.counter}"
                self.counter += 1
                request_responses.append(response)
            responses.append(Response(data=request_responses))
        return responses


class TestLocalAPI:
    TEST_TIMEOUT = 0.1  # Short timeout for tests

    @pytest.fixture
    def cache(self):
        """Provide a temporary cache for each test."""
        with tempfile.TemporaryDirectory() as tmpdir:
            with Cache(tmpdir) as cache:
                yield cache

    @pytest.mark.asyncio
    async def test_batch_size_optimization(self, cache):
        """Test that requests are properly batched before reaching model."""
        model = BatchTrackingModel()
        batch_size = 2

        async with LocalAPI(
            model=model,
            cache=cache,
            batch_size=batch_size,
            timeout=self.TEST_TIMEOUT
        ) as api:
            # Create 3 requests (should result in two batches: size 2 and 1)
            requests = [
                Request(args=(), kwargs={"messages": f"test{i}"}, n=1,
                        request_id=str(i), namespace="ns1")
                for i in range(3)
            ]

            # Send requests concurrently
            tasks = [api.request(req) for req in requests]
            await asyncio.gather(*tasks)

            # Verify batching
            assert len(model.batch_calls) == 2
            assert len(model.batch_calls[0].requests) == batch_size
            assert len(model.batch_calls[1].requests) == 1

    @pytest.mark.asyncio
    async def test_cache_reuse_across_namespaces(self, cache):
        """Test that responses are reused across namespaces while maintaining batch size."""
        model = BatchTrackingModel()
        batch_size = 2

        async with LocalAPI(
            model=model,
            cache=cache,
            batch_size=batch_size,
            timeout=self.TEST_TIMEOUT
        ) as api:
            # First namespace requests
            ns1_requests = [
                Request(args=(), kwargs={"messages": "test1"}, n=1,
                        request_id="1", namespace="ns1"),
                Request(args=(), kwargs={"messages": "test2"}, n=1,
                        request_id="2", namespace="ns1")
            ]
            ns1_results = await asyncio.gather(*[api.request(req) for req in ns1_requests])

            # Same prompts, different namespace
            ns2_requests = [
                Request(args=(), kwargs={"messages": "test1"}, n=1,
                        request_id="3", namespace="ns2"),
                Request(args=(), kwargs={"messages": "test2"}, n=1,
                        request_id="4", namespace="ns2")
            ]
            ns2_results = await asyncio.gather(*[api.request(req) for req in ns2_requests])

            # Verify cache reuse
            assert [r.data for r in ns1_results] == [r.data for r in ns2_results]  # Same responses reused
            assert len(model.batch_calls) == 1  # Only one batch call to model

    @pytest.mark.asyncio
    async def test_mixed_cache_and_fresh_requests(self, cache):
        """Test handling mix of cached and fresh requests while maintaining batch size."""
        model = BatchTrackingModel()
        batch_size = 2

        async with LocalAPI(
            model=model,
            cache=cache,
            batch_size=batch_size,
            timeout=self.TEST_TIMEOUT
        ) as api:
            # Initial request to populate cache
            request1 = Request(args=(), kwargs={"messages": "cached"}, n=1,
                               request_id="1", namespace="ns1")
            await api.request(request1)

            # Mix of cached and fresh requests
            mixed_requests = [
                Request(args=(), kwargs={"messages": "cached"}, n=1, request_id="2",
                        namespace="ns2"),  # Should use cache
                Request(args=(), kwargs={"messages": "fresh1"}, n=1, request_id="3",
                        namespace="ns1"),  # New request
                Request(args=(), kwargs={"messages": "fresh2"}, n=1, request_id="4",
                        namespace="ns1")   # New request
            ]

            results = await asyncio.gather(*[api.request(req) for req in mixed_requests])

            # Verify batching of fresh requests
            assert len(model.batch_calls) == 2
            assert len(model.batch_calls[1].requests) == 2
            assert all(req.kwargs.get('messages', '').startswith("fresh")
                       for req in model.batch_calls[1].requests)

    @pytest.mark.asyncio
    async def test_batch_request_interface(self, cache):
        """Test the batch_request interface directly."""
        model = BatchTrackingModel()
        batch_size = 2

        async with LocalAPI(
            model=model,
            cache=cache,
            batch_size=batch_size,
            timeout=self.TEST_TIMEOUT
        ) as api:
            batch = Batch(requests=[
                Request(args=(), kwargs={"messages": "test1"}, n=1,
                        request_id="1", namespace="ns1"),
                Request(args=(), kwargs={"messages": "test2"}, n=1,
                        request_id="2", namespace="ns1"),
                Request(args=(), kwargs={"messages": "test3"}, n=1,
                        request_id="3", namespace="ns1")
            ])

            results = await api.batch_request(batch)

            # Verify results structure
            assert len(results) == 3
            assert all(isinstance(r, Response) for r in results)

            # Verify batching
            assert len(model.batch_calls) == 2
            assert len(model.batch_calls[0].requests) == batch_size
            assert len(model.batch_calls[1].requests) == 1


class TestLocalAPIBatching:
    TEST_TIMEOUT = 0.1  # Short timeout for tests

    @pytest.fixture
    def cache(self):
        """Provide a temporary cache for each test."""
        with tempfile.TemporaryDirectory() as tmpdir:
            with Cache(tmpdir) as cache:
                yield cache

    @pytest.mark.asyncio
    async def test_allow_batch_overflow_single_large_request(self, cache):
        """Test overflow with a single request larger than batch size."""
        model = BatchTrackingModel()
        batch_size = 5

        async with LocalAPI(
            model=model,
            cache=cache,
            batch_size=batch_size,
            timeout=self.TEST_TIMEOUT,
            allow_batch_overflow=True
        ) as api:
            # Single request larger than batch size
            request1 = Request(args=(), kwargs={"messages": "large"}, n=7, request_id="r1", namespace="test")
            result1 = await api.request(request1)

            # Verify: Should process as one batch despite size
            assert len(result1.data) == 7
            assert len(model.batch_calls) == 1
            assert len(model.batch_calls[0].requests) == 7
    
    @pytest.mark.asyncio
    async def test_allow_batch_overflow_fills_existing(self, cache):
        """Test overflow where a request fills and exceeds existing items."""
        model = BatchTrackingModel()
        batch_size = 5

        async with LocalAPI(
            model=model,
            cache=cache,
            batch_size=batch_size,
            timeout=self.TEST_TIMEOUT,
            allow_batch_overflow=True
        ) as batcher:
            # Send 3 initial requests (batch size 5)
            initial_requests = [
                Request(args=(), kwargs={"messages": f"init{i}"}, n=1, request_id=f"i{i}", namespace="test")
                for i in range(3)
            ]
            replies = [batcher.request(req) for req in initial_requests]

            # Create extra request that leads to overflow
            overflow_request = Request(args=(), kwargs={"messages": "overflow"}, n=7, request_id="of1", namespace="test")
            overflow_reply = batcher.request(overflow_request)

            # Await the results
            replies.append(overflow_reply)
            all_results = await asyncio.gather(*replies)

            # Extract results for verification
            initial_results = all_results[:3]
            overflow_result = all_results[3]

            # Verify: Should still be one batch with all 10 requests
            assert len([r.data[0] for r in initial_results]) == 3
            assert len(overflow_result.data) == 7
            assert len(model.batch_calls) == 1
            assert len(model.batch_calls[0].requests) == 10 # 3 initial + 7 overflow



from contextlib import AbstractAsyncContextManager


class UnlimitedResource:
    def __init__(self, key):
        self.data = key

    def free(self, *args, **kwargs):
        pass


class UnlimitedLimiter(AbstractAsyncContextManager):

    def __init__(self, key):
        self.key = key

    async def __aenter__(self):
        return UnlimitedResource(self.key)

    async def __aexit__(self, exc_type, exc, tb):
        return None



from together import AsyncClient
from copy import deepcopy
from ..typedefs import Request, SingleRequestModel, Response
from .wrapper import AsyncWrapper


class TogetherAIModel(SingleRequestModel):
    def __init__(self):
        self.aclient = AsyncClient()

    async def request(self, request: Request):
        args = request.args
        kwargs = request.kwargs

        response = await self.aclient.chat.completions.create(
            *args, **kwargs, n=request.n
        )

        # ToDo: this is difficult, how do we deal with the case n>1
        # Cachesaver then expects a list of replies
        flattened_responses = []
        for choice in response.choices:
            response_copy = deepcopy(response)
            response_copy.choices = [choice]
            flattened_responses.append(response_copy)

        return Response(
            data=flattened_responses)


class AsyncTogether(AsyncWrapper):
    def __init__(
            self,
            namespace="default",
            cachedir="./cache",
            batch_size=1):
        
        model = TogetherAIModel()
        
        super().__init__(
            model=model,
            namespace=namespace,
            cachedir=cachedir,
            batch_size=batch_size)


import pytest
import logging
import tempfile
import asyncio
from typing import List, Any
from collections import defaultdict
from diskcache import Cache

from src.cachesaver.typedefs import Request, Response, Batch, BatchRequestModel
from src.cachesaver.pipelines import OrderedLocalAPI


class BatchCounterModel(BatchRequestModel):
    """Mock model that maintains per-prompt counters and tracks batch sizes."""

    def __init__(self):
        self.batch_calls: List[Batch] = []
        self.prompt_counters = defaultdict(int)

    async def batch_request(self, batch: Batch) -> List[List[Any]]:
        self.batch_calls.append(batch)
        responses = []
        for request in batch.requests:
            request_responses = []
            prompt = request.kwargs.get('messages', '')
            for _ in range(request.n):
                counter = self.prompt_counters[prompt]
                request_responses.append(f"{prompt}_count{counter}")
                self.prompt_counters[prompt] += 1
            responses.append(Response(data=request_responses))
        return responses


class TestOrderedLocalAPI:
    TEST_TIMEOUT = 0.1

    @pytest.fixture
    def cache(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            with Cache(tmpdir) as cache:
                yield cache

    @pytest.fixture(autouse=True)
    def setup_logging(self):
        """Configure logging for all tests in this module."""
        logging.basicConfig(
            level=logging.DEBUG,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )

    @pytest.mark.asyncio
    async def test_two_stage_batching(self, cache):
        """Test that requests go through both collection and hardware batching."""
        model = BatchCounterModel()

        async with OrderedLocalAPI(
            model=model,
            cache=cache,
            collection_batch_size=4,
            hardware_batch_size=2,
            timeout=self.TEST_TIMEOUT
        ) as api:
            requests = [
                Request(args=(), kwargs={"messages": f"p{i}"}, n=1,
                        request_id=str(i), namespace="ns1")
                for i in range(4)
            ]

            await asyncio.gather(*[api.request(req) for req in requests])

            # Should see 2 hardware batches (size 2 each)
            assert len(model.batch_calls) == 2
            assert len(model.batch_calls[0].requests) == 2
            assert len(model.batch_calls[1].requests) == 2

    @pytest.mark.asyncio
    async def test_ordering_with_cache(self, cache):
        """Test that responses maintain order even with cache hits."""
        model = BatchCounterModel()

        async with OrderedLocalAPI(
            model=model,
            cache=cache,
            collection_batch_size=3,
            hardware_batch_size=2,
            timeout=self.TEST_TIMEOUT
        ) as api:
            # First request to populate cache
            await api.request(
                Request(args=(), kwargs={"messages": "cached"}, n=1,
                        request_id="2", namespace="ns1")
            )

            # Mix of cached and new requests with out-of-order IDs
            requests = [
                Request(args=(), kwargs={"messages": "new"}, n=1,
                        request_id="3", namespace="ns1"),
                Request(args=(), kwargs={"messages": "cached"}, n=1,
                        request_id="1", namespace="ns2"),
                Request(args=(), kwargs={"messages": "new"}, n=1,
                        request_id="2", namespace="ns1")
            ]

            results = await asyncio.gather(*[api.request(req) for req in requests])

            # Results should be ordered by request_id regardless of cache status
            assert results[0].data[0].startswith("new_count")      # id "3"
            assert results[1].data[0].startswith(
                "cached_count")   # id "1" (from cache)
            assert results[2].data[0].startswith("new_count")      # id "2"

    @pytest.mark.asyncio
    async def test_ordering_within_hardware_batches(self, cache):
        """Test that ordering is maintained within hardware batches."""
        model = BatchCounterModel()

        async with OrderedLocalAPI(
            model=model,
            cache=cache,
            collection_batch_size=4,
            hardware_batch_size=2,
            timeout=self.TEST_TIMEOUT
        ) as api:
            batch = Batch(requests=[
                Request(args=(), kwargs={"messages": "p1"}, n=1,
                        request_id="4", namespace="ns1"),
                Request(args=(), kwargs={"messages": "p1"}, n=1,
                        request_id="2", namespace="ns1"),
                Request(args=(), kwargs={"messages": "p1"}, n=1,
                        request_id="3", namespace="ns1"),
                Request(args=(), kwargs={"messages": "p1"}, n=1,
                        request_id="1", namespace="ns1")
            ])

            results = await api.batch_request(batch)

            # Verify responses are ordered by request_id
            assert results[0].data[0] == "p1_count3"  # id "4"
            assert results[1].data[0] == "p1_count1"  # id "2"
            assert results[2].data[0] == "p1_count2"  # id "3"
            assert results[3].data[0] == "p1_count0"  # id "1"

            # Verify hardware batching
            assert len(model.batch_calls) == 2
            assert len(model.batch_calls[0].requests) == 2
            assert len(model.batch_calls[1].requests) == 2

    @pytest.mark.asyncio
    async def test_multiple_samples_ordering(self, cache):
        """Test ordering with requests for multiple samples."""
        model = BatchCounterModel()

        async with OrderedLocalAPI(
            model=model,
            cache=cache,
            collection_batch_size=3,
            hardware_batch_size=2,
            timeout=self.TEST_TIMEOUT
        ) as api:
            requests = [
                Request(args=(), kwargs={"messages": "p1"}, n=2,
                        request_id="2", namespace="ns1"),
                Request(args=(), kwargs={"messages": "p2"}, n=1,
                        request_id="1", namespace="ns1")
            ]

            results = await asyncio.gather(*[api.request(req) for req in requests])

            # Verify ordering and sample counts
            assert len(results[0].data) == 2  # n=2
            assert len(results[1].data) == 1  # n=1
            assert results[1].data[0].startswith("p2_count")  # id "1"
            assert all(r.startswith("p1_count") for r in results[0].data)  # id "2"


class TestLocalAPIBatching:
    TEST_TIMEOUT = 0.1  # Short timeout for tests

    @pytest.fixture
    def cache(self):
        """Provide a temporary cache for each test."""
        with tempfile.TemporaryDirectory() as tmpdir:
            with Cache(tmpdir) as cache:
                yield cache

    @pytest.mark.asyncio
    async def test_allow_batch_overflow_single_large_request(self, cache):
        """Test overflow with a single request larger than batch size."""
        model = BatchCounterModel()
        batch_size = 5

        async with OrderedLocalAPI(
            model=model,
            cache=cache,
            collection_batch_size=batch_size,
            hardware_batch_size=10, # Assuming hardware fits all requests
            timeout=self.TEST_TIMEOUT,
            allow_batch_overflow=True
        ) as api:
            # Single request larger than batch size
            request1 = Request(args=(), kwargs={"messages": "large"}, n=7, request_id="r1", namespace="test")
            result1 = await api.request(request1)

            # Verify: Should process as one batch despite size
            assert len(result1.data) == 7
            assert len(model.batch_calls) == 1
            assert len(model.batch_calls[0].requests) == 7
    
    @pytest.mark.asyncio
    async def test_allow_batch_overflow_fills_existing(self, cache):
        """Test overflow where a request fills and exceeds existing items."""
        model = BatchCounterModel()
        batch_size = 5

        async with OrderedLocalAPI(
            model=model,
            cache=cache,
            collection_batch_size=batch_size,
            hardware_batch_size=10, # Assuming hardware fits all requests
            timeout=self.TEST_TIMEOUT,
            allow_batch_overflow=True
        ) as batcher:
            # Send 3 initial requests (batch size 5)
            initial_requests = [
                Request(args=(), kwargs={"messages": f"init{i}"}, n=1, request_id=f"i{i}", namespace="test")
                for i in range(3)
            ]
            replies = [batcher.request(req) for req in initial_requests]

            # Create extra request that leads to overflow
            overflow_request = Request(args=(), kwargs={"messages": "overflow"}, n=7, request_id="of1", namespace="test")
            overflow_reply = batcher.request(overflow_request)

            # Await the results
            replies.append(overflow_reply)
            all_results = await asyncio.gather(*replies)

            # Extract results for verification
            initial_results = all_results[:3]
            overflow_result = all_results[3]

            # Verify: Should still be one batch with all 10 requests
            assert len([r.data[0] for r in initial_results]) == 3
            assert len(overflow_result.data) == 7
            assert len(model.batch_calls) == 1
            assert len(model.batch_calls[0].requests) == 10 # 3 initial + 7 overflow



import pytest
from typing import List, Any
from collections import defaultdict

from src.cachesaver.typedefs import Request, Batch, BatchRequestModel
from src.cachesaver.reordering import RequestReorderer


class PromptCounterModel(BatchRequestModel):
    """Mock model that maintains separate counters for each prompt."""

    def __init__(self):
        self.prompt_counters = defaultdict(int)
        self.calls: List[Batch] = []

    async def batch_request(self, batch: Batch) -> List[List[Any]]:
        self.calls.append(batch)
        responses = []
        for request in batch.requests:
            # Generate n responses for this request
            request_responses = []
            prompt = request.kwargs.get('messages', '')
            for _ in range(request.n):
                counter = self.prompt_counters[prompt]
                request_responses.append(f"{prompt}_count{counter}")
                self.prompt_counters[prompt] += 1
            responses.append(request_responses)
        return responses


class TestRequestReorderer:
    @pytest.mark.asyncio
    async def test_basic_reordering(self):
        """Test that responses are reordered according to request_ids."""
        model = PromptCounterModel()
        reorderer = RequestReorderer(model=model)

        batch = Batch(requests=[
            Request(args=(), kwargs={"messages": "p1"}, n=1,
                    request_id="2", namespace="test"),
            Request(args=(), kwargs={"messages": "p1"}, n=1,
                    request_id="1", namespace="test"),
            Request(args=(), kwargs={"messages": "p1"}, n=1,
                    request_id="3", namespace="test")
        ])

        results = await reorderer.batch_request(batch)

        # Verify the model received sorted requests
        received_batch = model.calls[0]
        assert [r.request_id for r in received_batch.requests] == ["1", "2", "3"]

        # Verify responses are in original order
        # request_id "2" was first in input
        assert results[0][0] == "p1_count1"
        # request_id "1" was second in input
        assert results[1][0] == "p1_count0"
        # request_id "3" was third in input
        assert results[2][0] == "p1_count2"

    @pytest.mark.asyncio
    async def test_different_prompts_reordering(self):
        """Test reordering with different prompts."""
        model = PromptCounterModel()
        reorderer = RequestReorderer(model=model)

        batch = Batch(requests=[
            Request(args=(), kwargs={"messages": "p2"}, n=1,
                    request_id="2", namespace="test"),
            Request(args=(), kwargs={"messages": "p1"}, n=1,
                    request_id="1", namespace="test"),
            Request(args=(), kwargs={"messages": "p2"}, n=1,
                    request_id="3", namespace="test")
        ])

        results = await reorderer.batch_request(batch)

        # Verify counters are separate per prompt
        assert results[0][0] == "p2_count0"  # First p2, request_id "2"
        assert results[1][0] == "p1_count0"  # First p1, request_id "1"
        assert results[2][0] == "p2_count1"  # Second p2, request_id "3"

    @pytest.mark.asyncio
    async def test_multiple_samples_reordering(self):
        """Test reordering with requests asking for multiple samples."""
        model = PromptCounterModel()
        reorderer = RequestReorderer(model=model)

        batch = Batch(requests=[
            Request(args=(), kwargs={"messages": "p1"}, n=2,
                    request_id="2", namespace="test"),
            Request(args=(), kwargs={"messages": "p2"}, n=1,
                    request_id="1", namespace="test")
        ])

        results = await reorderer.batch_request(batch)

        assert len(results[0]) == 2  # First request wanted 2 samples
        assert len(results[1]) == 1  # Second request wanted 1 sample
        assert results[0][0] == "p1_count0"  # First p1
        assert results[1][0] == "p2_count0"  # First p2

    @pytest.mark.asyncio
    async def test_lexicographic_ordering(self):
        """Test that request_ids are sorted lexicographically."""
        model = PromptCounterModel()
        reorderer = RequestReorderer(model=model)

        batch = Batch(requests=[
            Request(args=(), kwargs={"messages": "p1"}, n=1,
                    request_id="req10", namespace="test"),
            Request(args=(), kwargs={"messages": "p1"}, n=1,
                    request_id="req2", namespace="test")
        ])

        results = await reorderer.batch_request(batch)

        received_batch = model.calls[0]
        assert [r.request_id for r in received_batch.requests] == [
            "req10", "req2"]
        assert results[0][0] == "p1_count0"
        assert results[1][0] == "p1_count1"



import pytest
import asyncio
from typing import List, Any
from collections import defaultdict

from src.cachesaver.typedefs import Request, Response, Batch, BatchRequestModel
from src.cachesaver.batching import AsyncBatcher


class BatchTrackingModel(BatchRequestModel):
    """Model that tracks batch sizes and maintains per-prompt counters."""

    def __init__(self):
        self.batch_calls: List[Batch] = []
        self.prompt_counters = defaultdict(int)

    async def batch_request(self, batch: Batch) -> List[List[Any]]:
        self.batch_calls.append(batch)
        responses = []
        for request in batch.requests:
            prompt = request.kwargs.get('messages', '')
            counter = self.prompt_counters[prompt]
            responses.append(Response(data=[f"{prompt}_count{counter}"]))
            self.prompt_counters[prompt] += 1
        return responses


@pytest.mark.asyncio
async def test_stacked_batchers():
    """Test three stacked batchers with different batch sizes and timeouts."""
    model = BatchTrackingModel()

    async with AsyncBatcher(
        model=model,
        batch_size=2,
        timeout=0.05,
        name="inner"
    ) as inner_batcher:
        async with AsyncBatcher(
            model=inner_batcher,
            batch_size=4,
            timeout=0.1,
            name="middle"
        ) as middle_batcher:
            async with AsyncBatcher(
                model=middle_batcher,
                batch_size=3,
                timeout=0.03,
                name="outer"
            ) as outer_batcher:
                # Create test requests
                requests = [
                    Request(args=(), kwargs={"messages": f"prompt_{i}"},
                            n=1,
                            request_id=str(i),
                            namespace="test")
                    for i in range(10)
                ]

                # Send requests with delays
                tasks = []
                for request in requests:
                    task = asyncio.create_task(outer_batcher.request(request))
                    tasks.append(task)
                    await asyncio.sleep(0.04)  # Simulate gradual arrival

                # Wait for all requests
                results = await asyncio.gather(*tasks)

                # Verify results
                assert len(results) == 10
                assert all(isinstance(r, Response) for r in results)
                assert all(r.data[0].startswith("prompt_") for r in results)

                # Check batch sizes at model level
                batch_sizes = [len(batch.requests)
                               for batch in model.batch_calls]
                # Inner batcher limit
                assert all(size <= 2 for size in batch_sizes)



from typing import List, Any
from .typedefs import Request, Batch, SingleRequestModel, BatchRequestModel


class RequestReorderer(BatchRequestModel):
    """
    Ensures deterministic ordering of responses based on request_id.

    Requests are ordered by request_id before being sent to the model,
    and responses are reordered to match the original request order.
    """

    def __init__(self, model: BatchRequestModel):
        self.model = model

    async def batch_request(self, batch: Batch) -> List[List[Any]]:
        """Reorder batch requests by request_id and restore original order in responses."""
        # Create mapping of original positions
        assert len({req.request_id for req in batch.requests}) == len([req.request_id for req in batch.requests]),\
            f"Request IDs must be unique in a reorderer's batch: {[req.request_id for req in batch.requests]}"
        
        original_order = {req.request_id: i for i,
                          req in enumerate(batch.requests)}

        # Sort requests by request_id
        sorted_requests = sorted(batch.requests, key=lambda x: x.request_id)
        sorted_batch = Batch(requests=sorted_requests)

        # Get responses for sorted requests
        sorted_responses = await self.model.batch_request(sorted_batch)

        # Restore original order
        reordered_responses = [None] * len(sorted_responses)
        for req, resp in zip(sorted_requests, sorted_responses):
            original_idx = original_order[req.request_id]
            reordered_responses[original_idx] = resp

        return reordered_responses



import asyncio
import logging
from typing import List, Any, Dict
from diskcache import Cache
from .resources import AsyncResource

from .typedefs import Request, Batch, Response, SingleRequestModel, BatchRequestModel

logger = logging.getLogger(__name__)


class AsyncCacher(AsyncResource, SingleRequestModel, BatchRequestModel):
    """
    Caches responses from a model, with namespace-aware caching.

    While this class implements both SingleRequestModel and BatchRequestModel protocols,
    it requires a SingleRequestModel as input and handles batches by splitting them
    into individual requests.

    Within a namespace, each request gets fresh responses.
    Across namespaces, responses can be reused.
    """

    def __init__(self, model: SingleRequestModel, cache: Cache):
        assert cache is not None, "cache must be provided"
        assert model is not None, "model must be provided"
        assert isinstance(
            model, SingleRequestModel), "model must implement SingleRequestModel"

        self.model = model
        self.cache = cache
        self.key2mutex: Dict[str, asyncio.Lock] = {}
        self.namespace2used_counts: Dict[str, Dict[str, int]] = {}
        self.used_counts: Dict[str, int] = {}

    async def request(self, request: Request) -> Response:
        """Handle a single request, potentially reusing cached responses."""
        key = request.hash()

        # Get or create mutex for this prompt
        mutex = self.key2mutex.get(key, asyncio.Lock())
        self.key2mutex[key] = mutex

        async with mutex:
            entries_in_cache = self.cache.get(key, [])

            # Track usage counts per namespace
            if request.namespace is not None:
                used_counts = self.namespace2used_counts.get(
                    request.namespace, {})
                self.namespace2used_counts[request.namespace] = used_counts
            else:
                used_counts = self.used_counts

            used = used_counts.get(key, 0)

            # Check if we need more responses
            num_needed = max(request.n - len(entries_in_cache[used:]), 0)
            if num_needed > 0:
                # Get fresh responses from model
                response = await self.model.request(request)
                new_entries = response.data
                entries_in_cache.extend(new_entries)
                self.cache.set(key, entries_in_cache)

            # Update usage count
            used_counts[key] = used + request.n

            response = Response(
                data=entries_in_cache[used:used + request.n],
                cached=[True] * (request.n-num_needed) + [False] * num_needed,
                duplicated=[False] * request.n,
            )
            return response

    async def batch_request(self, batch: Batch) -> List[Response]:
        """Handle a batch of requests by processing each individually."""
        coroutines = [self.request(request) for request in batch.requests]
        return await asyncio.gather(*coroutines)

    async def cleanup(self):
        """Cleanup cache resources."""
        self.cache.close()



import asyncio
import time
import logging
from typing import List, Any
from dataclasses import replace
from .typedefs import Request, Batch, Response, SingleRequestModel, BatchRequestModel
from .resources import AsyncResource

logger = logging.getLogger(__name__)


class AsyncBatcher(AsyncResource, SingleRequestModel, BatchRequestModel):
    """
    Groups requests into batches using a queue-based producer-consumer pattern.
    Requests wait at most `timeout` seconds for the batch to be filled.
    If the batch is not filled in time, it is sent to the model as is.
    """
    SHUTDOWN_SIGNAL = (
        None, None, None)  # (timestamp, request, future) tuple for shutdown

    def __init__(self, model: BatchRequestModel, batch_size: int, timeout: float = 30,
                 name: str = "unnamed", allow_batch_overflow: bool = False):
        self.name = name
        self.batch_size = batch_size
        self.timeout = timeout
        self.model = model
        self.queue = asyncio.Queue()
        self.allow_batch_overflow = allow_batch_overflow

        # Start background worker
        self.worker_task = asyncio.create_task(self._batch_worker())
        logger.debug("[Batcher %s] Initialized with batch_size=%d, timeout=%f",
                     self.name, batch_size, timeout)

    async def cleanup(self):
        """Cleanup the background worker task."""
        logger.debug("[Batcher %s] Initiating shutdown", self.name)
        # Wake up worker with shutdown signal
        await self.queue.put(self.SHUTDOWN_SIGNAL)
        await self.worker_task
        logger.debug("[Batcher %s] Worker shut down cleanly", self.name)

    async def request(self, request: Request) -> List[Any]:
        """Handle a single request, potentially generating multiple samples."""
        logger.debug("[Batcher %s] Received request: %s", self.name, request)

        futures = []
        # Use put_nowait for all but the last request if num_desired > 1
        num_to_put_nowait = request.n - 1
        for i in range(request.n):
            single_request = replace(request, n=1, request_id=f"{request.request_id}_{i}")
            # Create future in current event loop
            future = asyncio.get_event_loop().create_future()
            timestamp = time.time()
            item_to_put = (timestamp, single_request, future)

            if i < num_to_put_nowait:
                try:
                    self.queue.put_nowait(item_to_put)
                    logger.debug(
                        "[Batcher %s] put_nowait item %s. Queue size: %d",
                        self.name, single_request.request_id, self.queue.qsize()
                    )
                except asyncio.QueueFull:
                    # If queue is full even with put_nowait, we have to wait
                    # This might break the "keep together" goal, but avoids losing requests
                    logger.warning(
                        "[Batcher %s] Queue full during put_nowait, falling back to await put.", self.name)
                    await self.queue.put(item_to_put)
                    logger.debug(
                        "[Batcher %s] await put item %s (fallback). Queue size: %d",
                        self.name, single_request.request_id, self.queue.qsize()
                    )
            else:
                # Await putting the last item (or the only item if num_desired == 1)
                await self.queue.put(item_to_put)
                logger.debug(
                    "[Batcher %s] await put item %s (last). Queue size: %d",
                    self.name, single_request.request_id, self.queue.qsize()
                )


            futures.append(future)

        # Wait for all samples
        responses = await asyncio.gather(*futures)
        response = Response(
            data=[r.data[0] for r in responses],
            cached=[r.cached[0] if r.cached is not None else False for r in responses],
            duplicated=[r.duplicated[0] if r.duplicated is not None else False for r in responses] 
        )
        return response  # Unwrap single-item responses

    async def batch_request(self, batch: Batch) -> List[List[Any]]:
        """Process a batch of requests."""
        return await asyncio.gather(*[self.request(req) for req in batch.requests])

    async def _batch_worker(self):
        """Background worker that processes requests in batches."""
        try:
            while True:
                batch = []
                futures = []
                requests = []

                try:
                    # Get first request to start the batch
                    first_item = await self.queue.get()
                    logger.debug(
                        "[Batcher %s] await get first item. Queue size: %d",
                        self.name, self.queue.qsize()
                    )

                    # Check for shutdown signal
                    if first_item == self.SHUTDOWN_SIGNAL:
                        logger.debug(
                            "[Batcher %s] Received shutdown signal", self.name)
                        break

                    first_timestamp, first_request, first_future = first_item
                    batch.append(first_request)
                    futures.append(first_future)
                    requests.append(first_request)

                    # Collect more requests up to batch_size or timeout
                    while len(batch) < self.batch_size:
                        now = time.time()
                        max_wait = max(0, self.timeout -
                                       (now - first_timestamp))

                        try:
                            item = await asyncio.wait_for(
                                self.queue.get(), timeout=max_wait)
                            logger.debug(
                                "[Batcher %s] await get subsequent item. Queue size: %d",
                                self.name, self.queue.qsize()
                            )

                            # Check for shutdown signal
                            if item == self.SHUTDOWN_SIGNAL:
                                logger.debug("[Batcher %s] Received shutdown signal mid-batch",
                                             self.name)
                                # Process current batch before shutting down
                                break

                            timestamp, request, future = item
                            batch.append(request)
                            futures.append(future)
                            requests.append(request)
                        except asyncio.TimeoutError:
                            logger.debug("[Batcher %s] Timeout waiting for batch completion",
                                         self.name)
                            break
                    
                    # Add overflow items if allowed
                    if self.allow_batch_overflow:
                        logger.debug(
                            "[Batcher %s] Checking for overflow items.", self.name)
                        try:
                            while True:  # Drain queue without waiting
                                item = self.queue.get_nowait()
                                logger.debug(
                                    "[Batcher %s] get_nowait item in overflow. Queue size: %d",
                                    self.name, self.queue.qsize()
                                )
                                if item == self.SHUTDOWN_SIGNAL:
                                    # If shutdown received during drain, just stop draining.
                                    # The main loop will handle the signal again.
                                    logger.debug(
                                        "[Batcher %s] Shutdown signal hit during overflow drain.", self.name)
                                    break
                                timestamp, request, future = item
                                batch.append(request)
                                futures.append(future)
                                requests.append(request)
                                logger.debug(
                                    "[Batcher %s] Added overflow item %s.", self.name, request.request_id)
                        except asyncio.QueueEmpty:
                            pass  # No more items available immediately

                    # Add overflow items if allowed
                    if self.allow_batch_overflow:
                        logger.debug(
                            "[Batcher %s] Checking for overflow items.", self.name)
                        try:
                            while True:  # Drain queue without waiting
                                item = self.queue.get_nowait()
                                logger.debug(
                                    "[Batcher %s] get_nowait item in overflow. Queue size: %d",
                                    self.name, self.queue.qsize()
                                )
                                if item == self.SHUTDOWN_SIGNAL:
                                    # If shutdown received during drain, just stop draining.
                                    # The main loop will handle the signal again.
                                    logger.debug(
                                        "[Batcher %s] Shutdown signal hit during overflow drain.", self.name)
                                    break
                                timestamp, request, future = item
                                batch.append(request)
                                futures.append(future)
                                requests.append(request)
                                logger.debug(
                                    "[Batcher %s] Added overflow item %s.", self.name, request.request_id)
                        except asyncio.QueueEmpty:
                            pass  # No more items available immediately

                    # Process batch if we have any requests
                    if batch:
                        try:
                            logger.debug("[Batcher %s] Processing batch of size %d",
                                         self.name, len(batch))
                            responses = await self.model.batch_request(Batch(requests=requests))
                            for future, response in zip(futures, responses):
                                if not future.done():
                                    future.set_result(response)
                        except Exception as e:
                            logger.error("[Batcher %s] Batch processing failed: %s",
                                         self.name, str(e))
                            for future in futures:
                                if not future.done():
                                    future.set_exception(e)

                except Exception as e:
                    logger.error("[Batcher %s] Error processing batch: %s",
                                 self.name, str(e))
                    for future in futures:
                        if not future.done():
                            future.set_exception(e)

        except Exception as e:
            logger.error("[Batcher %s] Worker failed with unexpected error: %s",
                         self.name, str(e))
            # Make sure to cancel any pending futures before exiting
            for future in futures:
                if not future.done():
                    future.set_exception(e)
            raise
        finally:
            # Cancel any remaining futures when shutting down
            while not self.queue.empty():
                item = await self.queue.get()
                if item == self.SHUTDOWN_SIGNAL:
                    continue
                _, _, future = item
                if not future.done():
                    future.cancel()
            logger.debug(
                "[Batcher %s] Worker shut down, all futures resolved", self.name)
